import types
from collections import abc
from functools import partial
from inspect import signature
from typing import Dict, Callable, TypeVar, List, Any

from google.protobuf import json_format

Package = TypeVar('Package')


class StreamMessage(Dict):
    raw_data = None


def stream_request(request_iterator):
    for req in request_iterator:
        msg = StreamMessage(**json_format.MessageToDict(req))
        msg.raw_data = req
        yield msg


def parse_to_dict(item, message):
    return json_format.ParseDict(item, message) if isinstance(item, dict) else item


def parse_stream_return(results, message):
    for item in results:
        yield parse_to_dict(item, message)


class GrpcService:
    def __init__(self, pkg: Package, name: str, pb2: None):
        self.name = name
        self.pb2 = pb2
        self._descriptor = None
        self.add_service_func = getattr(pkg, f"add_{self.name}Servicer_to_server")
        self.__methods = {}

    def add_method(self, name: str, func: Callable):
        self.__methods[name] = func

    def warp_handler(self, func, message):
        sig = signature(func)
        parameters = [k for k, v in sig._parameters.items() if v.kind.value == 1]
        print(parameters)

        def decorator(self, request, context):
            # parse request
            if isinstance(request, abc.Iterable):  # request stream
                result = func(stream_request(request), context)
            else:  # request unary
                request_dict = json_format.MessageToDict(request)
                args = {}
                for p in parameters:
                    args[p] = request_dict.get(p, None)

                result = func(**args, request=request, context=context)

            # parse response
            if isinstance(result, types.GeneratorType):  # return stream
                return parse_stream_return(result, message)
            else:  # return anary
                return parse_to_dict(result, message)

        return decorator

    @property
    def descriptor(self):
        if not self._descriptor:
            self._descriptor = self.pb2.DESCRIPTOR.services_by_name[self.name]
        return self._descriptor

    def get_response_msg_type(self, name):
        return self.descriptor.methods_by_name[name].output_type._concrete_class

    def make_servicer(self):
        methods = {
            k: self.warp_handler(func, self.get_response_msg_type(k)())
            for k, func in self.__methods.items()
        }
        servicer = type(self.name, (), methods)
        return servicer()

    def add_to_server(self, server):
        self.add_service_func(self.make_servicer(), server)


def find_pb2_module(pkg) -> List[types.ModuleType]:
    return [v for k, v in pkg.__dict__.items() if isinstance(v, types.ModuleType) and hasattr(v, 'DESCRIPTOR')]


def find_module_by_has_service_name(modules: List[types.ModuleType], name: str) -> types.ModuleType:
    for module in modules:
        if module.DESCRIPTOR.services_by_name.get(name):
            return module


class GrpcServiceRegister:
    def __init__(self):
        self.files: Dict[Package, Dict[str, GrpcService]] = {}
        self.pkg_pb2: Dict[Package, List[types.ModuleType]] = {}

    def add(self, pkg, service, method, func):
        print(pkg, service, method)
        if not self.files.get(pkg):
            self.files[pkg] = {}
            self.pkg_pb2[pkg] = find_pb2_module(pkg)
        if not self.files[pkg].get(service):
            pb2 = find_module_by_has_service_name(self.pkg_pb2[pkg], service)
            self.files[pkg][service] = GrpcService(pkg, service, pb2)

        self.files[pkg][service].add_method(method, func)

    def add_to_server(self, server):
        for services in self.files.values():
            for service in services.values():
                service.add_to_server(server)

    @property
    def services(self) -> List[Any]:
        svcs = []
        for module, services in self.files.items():
            for name, service in services.items():
                svcs.append(service.descriptor)
        return svcs

    @property
    def test_servicers(self) -> Dict[Any, Any]:  # for testcase
        servicers = {}
        for module, services in self.files.items():

            for name, service in services.items():
                servicers[service.descriptor] = service.make_servicer()
        return servicers

    @property
    def service_names(self) -> List[str]:
        return [svc.full_name for svc in self.services]


_default_registry = None


def get_default_registry() -> GrpcServiceRegister:
    """
    Get the default registry to be used by the decorators and the reactor
    unless the explicit registry is provided to them.
    """
    global _default_registry
    if _default_registry is None:
        _default_registry = GrpcServiceRegister()
    return _default_registry


class WrapHandler:
    def __init__(self, pkg, service, method, func: Callable = None):
        self.func = func
        self.method = method or self.func.__name__
        get_default_registry().add(pkg, service, self.method, self.func)


def register(pkg, service, method=None):
    return partial(WrapHandler, pkg, service, method)



from concurrent import futures

import grpc
from grpc_reflection.v1alpha import reflection

from .registries import get_default_registry


class Server:
    def __init__(self, host: str = '127.0.0.1', port: str = '50051', worker: int = 10):
        self.host = host
        self.port = port
        self.worker = worker
        self.load_config_from_env()
        self.server = None

    def load_config_from_env(self):
        pass

    def run(self):
        registry = get_default_registry()
        self.server = grpc.server(futures.ThreadPoolExecutor(max_workers=self.worker))
        registry.add_to_server(self.server)
        SERVICE_NAMES = (
            *registry.service_names,
            reflection.SERVICE_NAME,
        )
        reflection.enable_server_reflection(SERVICE_NAMES, self.server)
        self.server.add_insecure_port(f'[::]:{self.port}')
        self.server.start()
        print('run server')
        print(f'# port : {self.port}')
        self.server.wait_for_termination()



import sys
from os.path import dirname

import click


@click.group()
def cli():
    pass


@click.command("run", short_help="Run a development server.")
@click.argument('file', type=click.Path(exists=True, resolve_path=True), default="app.py")
@click.option("--host", "-h", default="127.0.0.1", help="The interface to bind to.")
@click.option("--port", "-p", default='50051', help="The port to bind to.")
@click.option('--worker', '-w', default=10, type=int)
# @click.option(
#     "--reload/--no-reload",
#     default=None,
#     help="Enable or disable the reloader. By default the reloader "
#          "is active if debug is enabled.",
# )
@click.option(
    "--debug",
    default=False,
    help="Server Debug Mode",
)
def run_command(file, host, port, worker, debug):
    sys.path.append(dirname(file))
    import importlib.util

    spec = importlib.util.spec_from_file_location("app", file)
    foo = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(foo)

    from .server import Server
    Server(host, port, worker).run()


cli.add_command(run_command)

if __name__ == '__main__':
    cli()



import re

from setuptools import setup

with open("src/homi/__init__.py", encoding="utf8") as f:
    version = re.search(r'__version__ = "(.*?)"', f.read()).group(1)

setup(
    name='homi',
    version=version,
    install_requires=[
        "grpcio>=1.30.0",
        "grpcio_testing>=1.30.0",
        "grpcio-reflection>=1.30.0",
        "google-api-core>=1.21.0",
        "click>=7.1.2"
    ],
)



from . import helloworld_pb2_grpc
from ... import homi


# unary-unary method
@homi.register(helloworld_pb2_grpc, 'Greeter')
def SayHello(name, **kwargs):
    return {"message": f"Hello {name}!"}


# unary-stream method
@homi.register(helloworld_pb2_grpc, 'Greeter')
def SayHelloGroup(name, **kwargs):
    names = ['a', 'b', 'c', 'd']
    for name in names:
        yield {"message": f"Hello {name}!"}


# stream-unary method
@homi.register(helloworld_pb2_grpc, 'Greeter')
def HelloEveryone(request_iterator, context):
    names = []
    for reqs in request_iterator:
        names.append(reqs['name'])
    return {"message": f"Hello everyone {names}!"}


# stream-stream method
@homi.register(helloworld_pb2_grpc, 'Greeter')
def SayHelloOneByOne(request_iterator, context):
    for req in request_iterator:
        yield {"message": f"Hello {req['name']}!"}



import unittest

import grpc_testing

from .registries import get_default_registry


class HomiTestCase(unittest.TestCase):
    app = None
    _test_server = None

    def get_test_server(self):
        if not self._test_server:
            register = get_default_registry()
            self._test_server = grpc_testing.server_from_dictionary(
                register.test_servicers, grpc_testing.strict_real_time())

        return self._test_server

    @staticmethod
    def get_all_response(method):
        finish = False
        result = []
        while not finish:
            try:
                result.append(method.take_response())
            except Exception:
                finish = True
        return result

    @staticmethod
    def send_request_all(method, requests):
        [method.send_request(req) for req in requests]
        method.requests_closed()



import unittest

import grpc

from . import app, helloworld_pb2
from .helloworld_pb2 import HelloRequest
from ...homi import HomiTestCase


class GreeterTestCase(HomiTestCase):
    app = app

    def test_hello_say(self):
        server = self.get_test_server()
        name = "tom"
        request = HelloRequest(name=name)
        method = server.invoke_unary_unary(
            method_descriptor=(helloworld_pb2.DESCRIPTOR.services_by_name['Greeter'].methods_by_name['SayHello']),
            invocation_metadata={},
            request=request, timeout=1)

        response, metadata, code, details = method.termination()
        self.assertEqual(response.message, f'Hello {name}!')
        self.assertEqual(code, grpc.StatusCode.OK)

    def test_hello_say_group(self):
        server = self.get_test_server()
        name = "groupA"
        request = HelloRequest(name=name)
        method = server.invoke_unary_stream(
            method_descriptor=(helloworld_pb2.DESCRIPTOR.services_by_name['Greeter'].methods_by_name['SayHelloGroup']),
            invocation_metadata={},
            request=request, timeout=1)

        reps = self.get_all_response(method)

        self.assertEqual(len(reps), 4)
        return_names = ['a', 'b', 'c', 'd']
        for idx, rep in enumerate(reps):
            self.assertEqual(rep.message, f"Hello {return_names[idx]}!")

        metadata, code, details = method.termination()
        self.assertEqual(code, grpc.StatusCode.OK)

    def test_hello_everyone(self):
        server = self.get_test_server()
        method = server.invoke_stream_unary(
            method_descriptor=(helloworld_pb2.DESCRIPTOR.services_by_name['Greeter'].methods_by_name['HelloEveryone']),
            invocation_metadata={},
            timeout=1
        )

        names = ["tom", 'sam', 'wony', 'homi']
        self.send_request_all(method, (HelloRequest(name=name) for name in names))

        response, metadata, code, details = method.termination()
        self.assertEqual(code, grpc.StatusCode.OK)
        self.assertEqual(response.message, f'Hello everyone {names}!')

    def test_say_hello_one_by_one(self):
        server = self.get_test_server()
        method = server.invoke_stream_stream(
            method_descriptor=(
                helloworld_pb2.DESCRIPTOR.services_by_name['Greeter'].methods_by_name['SayHelloOneByOne']),
            invocation_metadata={},
            timeout=1
        )

        names = ["tom", 'sam', 'wony', 'homi']
        self.send_request_all(method, (HelloRequest(name=name) for name in names))

        reps = self.get_all_response(method)
        self.assertEqual(len(reps), len(names))

        for rep, name in zip(reps, names):
            self.assertEqual(rep.message, f'Hello {name}!')

        metadata, code, details = method.termination()
        self.assertEqual(code, grpc.StatusCode.OK)


if __name__ == '__main__':
    unittest.main()
